###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from PyQt4 import QtCore, QtGui
from itertools import izip
import os
import string

from vistrails.core import debug
from vistrails.core.configuration import get_vistrails_configuration
from vistrails.core.modules.basic_modules import identifier as basic_identifier
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.core.modules.utils import create_port_spec_string
from vistrails.core.vistrail.port_spec import PortSpec
from vistrails.core.system import vistrails_root_directory
from vistrails.gui.modules.utils import get_widget_class
from vistrails.gui.common_widgets import QToolWindowInterface
from vistrails.gui.port_documentation import QPortDocumentation
from vistrails.gui.theme import CurrentTheme

def letterIcon(letter, crossed=False):
    """ Creates icon with a specific letter
    """
    pixmap = QtGui.QPixmap(48,48)
    pixmap.fill(QtCore.Qt.transparent)
    painter = QtGui.QPainter(pixmap)
    painter.setPen(QtGui.QColor(0, 0, 0, 255))
    font = painter.font()
    font.setPointSize(40)
    painter.setFont(font)
    painter.drawText(0, 0, 48, 48, QtCore.Qt.AlignCenter, letter)
    if crossed:
        painter.drawLine(0,0,48,48)
        painter.drawLine(0,48,48,0)
    painter.end()
    return QtGui.QIcon(pixmap)

class AliasLabel(QtGui.QLabel):
    """
    AliasLabel is a QLabel that supports hover actions similar
    to a hot link
    """
    def __init__(self, alias='', text='', default_label='', parent=None):
        """ AliasLabel(alias:str , text: str, default_label: str,
                             parent: QWidget) -> QHoverAliasLabel
        Initialize the label with a text
        
        """
        QtGui.QLabel.__init__(self, parent)
        self.alias = alias
        self.caption = text
        # catch None
        if default_label:
            self.default_label = default_label
        else:
            self.default_label = ""
        self.updateText()
        self.setAttribute(QtCore.Qt.WA_Hover)
        self.setCursor(QtCore.Qt.PointingHandCursor)
        self.setToolTip(alias)
        self.palette().setColor(QtGui.QPalette.WindowText,
                                CurrentTheme.HOVER_DEFAULT_COLOR)

    def updateText(self):
        """ updateText() -> None
        Update the label text to contain the alias name when appropriate
        
        """
        if self.alias != '':
            self.setText(self.alias + ': ' + self.caption)
        elif self.default_label != '':
            self.setText(self.default_label + ': ' + self.caption)
        else:
            self.setText(self.caption)

    def event(self, event):
        """ event(event: QEvent) -> Event Result
        Override to handle hover enter and leave events for hot links
        
        """
        if event.type()==QtCore.QEvent.HoverEnter:
            self.palette().setColor(QtGui.QPalette.WindowText,
                                    CurrentTheme.HOVER_SELECT_COLOR)
        if event.type()==QtCore.QEvent.HoverLeave:
            self.palette().setColor(QtGui.QPalette.WindowText,
                                    CurrentTheme.HOVER_DEFAULT_COLOR)
        return QtGui.QLabel.event(self, event)
        # return super(QHoverAliasLabel, self).event(event)

    def mousePressEvent(self, event):
        """ mousePressEvent(event: QMouseEvent) -> None        
        If mouse click on the label, show up a dialog to change/add
        the alias name
        
        """
        if event.button()==QtCore.Qt.LeftButton:
            (text, ok) = QtGui.QInputDialog.getText(self,
                                                    'Set Parameter Alias',
                                                    'Enter the parameter alias',
                                                    QtGui.QLineEdit.Normal,
                                                    self.alias)
            while ok and self.parent().check_alias(str(text)):
                msg =" This alias is already being used.\
 Please enter a different parameter alias "
                (text, ok) = QtGui.QInputDialog.getText(self,
                                                        'Set Parameter Alias',
                                                        msg,
                                                        QtGui.QLineEdit.Normal,
                                                        text)
            if ok and str(text)!=self.alias:
                if not self.parent().check_alias(str(text)):
                    self.alias = str(text).strip()
                    self.updateText()
                    self.parent().updateMethod()

class Parameter(object):
    def __init__(self, desc, psi=None):
        self.type = desc.name
        self.identifier = desc.identifier
        self.namespace = None if not desc.namespace else desc.namespace
        self.strValue = ''
        self.alias = ''
        self.queryMethod = None
        self.port_spec_item = psi
        self.param_exists = False

class Function(object):
    def __init__(self, name, params, port_spec=None):
        self.name = name
        self.parameters = params
        self.port_spec = port_spec

    def get_spec(self, port_type):
        return self.port_spec

        
class ParameterEntry(QtGui.QTreeWidgetItem):
    plus_icon = QtGui.QIcon(os.path.join(vistrails_root_directory(),
                                         'gui/resources/images/plus.png'))
    minus_icon = QtGui.QIcon(os.path.join(vistrails_root_directory(),
                                          'gui/resources/images/minus.png'))
    def __init__(self, port_spec, function=None, types_visible=True, parent=None):
        QtGui.QTreeWidgetItem.__init__(self, parent)
        self.setFirstColumnSpanned(True)
        self.port_spec = port_spec
        self.function = function
        self.types_visible = types_visible

    def build_widget(self, widget_accessor, with_alias=True):
        reg = get_module_registry()

        # widget = QtGui.QDockWidget()
        # widget.setFeatures(QtGui.QDockWidget.DockWidgetClosable |
        #                    QtGui.QDockWidget.DockWidgetVerticalTitleBar)
        widget = QtGui.QWidget()

        h_layout = QtGui.QHBoxLayout()
        h_layout.insertSpacing(0, 10)
        h_layout.setMargin(2)
        h_layout.setSpacing(2)

        v_layout = QtGui.QVBoxLayout()
        v_layout.setAlignment(QtCore.Qt.AlignVCenter)
        delete_button = QtGui.QToolButton()
        delete_button.setIconSize(QtCore.QSize(8,8))
        delete_button.setIcon(ParameterEntry.minus_icon)
        def delete_method():
            if self.function is not None:
                self.group_box.parent().parent().parent().delete_method(
                    self, self.port_spec.name, self.function.real_id)
            else:
                self.group_box.parent().parent().parent().delete_method(
                    self, self.port_spec.name, None)
                
        QtCore.QObject.connect(delete_button, QtCore.SIGNAL("clicked()"), 
                               delete_method)
        v_layout.addWidget(delete_button)
        
        add_button = QtGui.QToolButton()
        add_button.setIcon(ParameterEntry.plus_icon)
        add_button.setIconSize(QtCore.QSize(8,8))
        def add_method():
            self.group_box.parent().parent().parent().add_method(
                self.port_spec.name)
        QtCore.QObject.connect(add_button, QtCore.SIGNAL("clicked()"), 
                               add_method)
        v_layout.addWidget(add_button)
        h_layout.addLayout(v_layout)
        
        self.my_widgets = []
        self.my_labels = []
        self.group_box = QtGui.QGroupBox()
        layout = QtGui.QGridLayout()
        layout.setMargin(5)
        layout.setSpacing(5)
        layout.setColumnStretch(1,1)
        self.group_box.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.group_box.setSizePolicy(QtGui.QSizePolicy.Preferred,
                                     QtGui.QSizePolicy.Fixed)
        self.group_box.palette().setColor(QtGui.QPalette.Window,
                                     CurrentTheme.METHOD_SELECT_COLOR)

        if self.function is not None:
            params = self.function.parameters
        else:
            params = [None,] * len(self.port_spec.descriptors())

        for i, (psi, param) in enumerate(izip(self.port_spec.port_spec_items, 
                                              params)):
            if psi.entry_type is not None:
                # !!only pull off the prefix!! options follow in camelcase
                prefix_end = len(psi.entry_type.lstrip(string.lowercase))
                if prefix_end == 0:
                    entry_type = psi.entry_type
                else:
                    entry_type = psi.entry_type[:-prefix_end]
            else:
                entry_type = None
            widget_class = widget_accessor(psi.descriptor, entry_type)
            if param is not None:
                obj = param
            else:
                obj = Parameter(psi.descriptor)
            obj.port_spec_item = psi

            if self.types_visible:
                if with_alias:
                    label = AliasLabel(obj.alias, obj.type, psi.label)
                    self.my_labels.append(label)
                else:
                    label = QtGui.QLabel(obj.type)
                layout.addWidget(label, i, 0)
                layout.setAlignment(label, QtCore.Qt.AlignLeft)

            param_widget = widget_class(obj, self.group_box)
            self.my_widgets.append(param_widget)
            layout.addWidget(param_widget, i, 1)
            layout.addItem(QtGui.QSpacerItem(0,0, QtGui.QSizePolicy.MinimumExpanding), i, 2)

        self.group_box.setLayout(layout)
        def updateMethod():
            if self.function is not None:
                real_id = self.function.real_id
            else:
                real_id = -1
            self.group_box.parent().parent().parent().update_method(
                self, self.port_spec.name, self.my_widgets, self.my_labels, real_id)
        def check_alias(name):
            controller = self.group_box.parent().parent().parent().controller
            if controller:
                return controller.check_alias(name)
            return False
        self.group_box.updateMethod = updateMethod
        self.group_box.check_alias = check_alias
        h_layout.addWidget(self.group_box)
        widget.setLayout(h_layout)
        return widget

    def get_widget(self):
        return self.build_widget(get_widget_class, True)

class PortItem(QtGui.QTreeWidgetItem):
    edit_show =  QtGui.QIcon(os.path.join(vistrails_root_directory(),
                             'gui/resources/images/pencil.png'))
    edit_hide = QtGui.QIcon(os.path.join(vistrails_root_directory(),
                             'gui/resources/images/pencil-disabled.png'))
    eye_open_icon = \
        QtGui.QIcon(os.path.join(vistrails_root_directory(),
                                 'gui/resources/images/eye.png'))
    eye_closed_icon = \
        QtGui.QIcon(os.path.join(vistrails_root_directory(),
                                 'gui/resources/images/eye_closed.png'))
    eye_disabled_icon = \
        QtGui.QIcon(os.path.join(vistrails_root_directory(),
                                 'gui/resources/images/eye_gray.png'))
    conn_icon = \
        QtGui.QIcon(os.path.join(vistrails_root_directory(),
                                 'gui/resources/images/connection.png'))

    def __init__(self, port_spec, is_connected, is_optional, is_visible,
                 is_editable=False, parent=None):
        QtGui.QTreeWidgetItem.__init__(self, parent)
        # self.setFlags(QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsEnabled)
        self.setFlags(QtCore.Qt.ItemIsEnabled)
        # self.setCheckState(0, QtCore.Qt.Unchecked)
        self.port_spec = port_spec
        self.is_connected = is_connected
        self.is_optional = is_optional
        self.is_visible = is_visible
        self.is_editable = is_editable
        self.is_unset = False
        self.build_item(port_spec, is_connected, is_optional, is_visible,
                        is_editable)

    def visible(self):
        return not self.is_optional or self.is_visible

    def set_visible(self, visible):
        self.is_visible = visible
        if visible:
            self.setIcon(0, PortItem.eye_open_icon)
        else:
            self.setIcon(0, PortItem.eye_closed_icon)

    def set_editable(self, edit):
        self.is_editable = edit
        if edit:
            self.setIcon(0, PortItem.edit_show)
        else:
            self.setIcon(0, PortItem.edit_hide)

    def get_visible(self):
        return self.visible_checkbox

    def get_connected(self):
        return self.connected_checkbox

    def is_constant(self):
        return (self.port_spec.is_valid and 
                get_module_registry().is_constant(self.port_spec))

    def calcUnset(self):
        self.is_unset = self.is_constant() and \
                        self.port_spec.is_mandatory() and \
                        not self.is_connected and \
                        not self.isExpanded()
        if self.is_unset:
            font = self.font(3)
            font.setWeight(QtGui.QFont.Bold)
            self.setFont(3, font)

    def build_item(self, port_spec, is_connected, is_optional, is_visible, is_editable):
        if not is_optional:
            self.setIcon(1, PortItem.eye_disabled_icon)
        elif is_visible:
            self.setIcon(1, PortItem.eye_open_icon)
        else:
            self.setIcon(1, PortItem.eye_closed_icon)

        if is_connected:
            self.setIcon(2, PortItem.conn_icon)
        self.setText(3, port_spec.name)

        if self.is_constant():
            if len(self.port_spec.port_spec_items)>0:
                if is_editable:
                    self.setIcon(0, PortItem.edit_show)
                else:
                    self.setIcon(0, PortItem.edit_hide)
        else:
            # if port_spec is not a method, make it gray
            self.setForeground(3,
                               QtGui.QBrush(QtGui.QColor.fromRgb(128,128,128)))

        self.visible_checkbox = QtGui.QCheckBox()
        self.connected_checkbox = QtGui.QCheckBox()
        
    def contextMenuEvent(self, event, widget):
        if self.port_spec is None:
            return
        act = QtGui.QAction("View Documentation", widget)
        act.setStatusTip("View method documentation")
        QtCore.QObject.connect(act,
                               QtCore.SIGNAL("triggered()"),
                               self.view_documentation)
        menu = QtGui.QMenu(widget)
        menu.addAction(act)
        menu.exec_(event.globalPos())
        
    def view_documentation(self):
        # descriptor = self.treeWidget().module.module_descriptor
        module = self.treeWidget().module
        port_type = self.treeWidget().port_type
        widget = QPortDocumentation(module,
                                    port_type,
                                    self.port_spec.name)
        widget.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        widget.exec_()

    def __lt__(self, other):
        # put unset mandatory ports first
        if self.is_unset != other.is_unset:
            return self.is_unset and not other.is_unset
        # put set (expanded) functions first
        if self.isExpanded() != other.isExpanded():
            return self.isExpanded() and not other.isExpanded()
        # otherwise use port name
        return self.port_spec.name < other.port_spec.name

class PortsList(QtGui.QTreeWidget):
    def __init__(self, port_type, parent=None):
        QtGui.QTreeWidget.__init__(self, parent)
        self.port_type = port_type
        self.setColumnCount(4)
        self.setColumnWidth(0,24)
        self.setColumnWidth(1,24)
        self.setColumnWidth(2,24)
        self.setRootIsDecorated(False)
        self.setIndentation(0)
        self.setHeaderHidden(True)
        self.connect(self, QtCore.SIGNAL("itemClicked(QTreeWidgetItem*, int)"),
                     self.item_clicked)
        self.module = None
        self.port_spec_items = {}
        self.entry_klass = ParameterEntry
        self.ports_visible = True
        self.types_visible = True

    def setReadOnly(self, read_only):
        self.setEnabled(not read_only)

    def set_entry_klass(self, entry_klass):
        if entry_klass != entry_klass:
            self.entry_klass = entry_klass
            self.update_module(self.module)

    def update_module(self, module):
        """ update_module(module: Module) -> None        
        Setup this tree widget to show functions of module
        
        """
        self.setColumnHidden(0, True)
        self.setColumnHidden(1, not self.ports_visible)
        # this is strange but if you try to clear the widget when the focus is 
        # in one of the items (after setting a parameter for example), 
        # VisTrails crashes on a Mac (Emanuele) This is probably a Qt bug
        w =  QtGui.QApplication.focusWidget()
        if self.isAncestorOf(w):
            w.clearFocus()
        self.clear()
        self.module = module
        self.port_spec_items = {}
        self.function_map = {}
        if module and module.is_valid:
            reg = get_module_registry()
            descriptor = module.module_descriptor
            if self.port_type == 'input':
                self.setColumnHidden(0,not get_vistrails_configuration(
                                        ).check('showInlineParameterWidgets'))
                port_specs = module.destinationPorts()
                connected_ports = module.connected_input_ports
                visible_ports = module.visible_input_ports
            elif self.port_type == 'output':
                port_specs = module.sourcePorts()
                connected_ports = module.connected_output_ports
                visible_ports = module.visible_output_ports
            else:
                raise TypeError("Unknown port type: '%s'" % self.port_type)
            
            for port_spec in sorted(port_specs, key=lambda x: x.name):
                connected = port_spec.name in connected_ports and \
                    connected_ports[port_spec.name] > 0
                item = PortItem(port_spec,
                                connected,
                                port_spec.optional,
                                port_spec.name in visible_ports,
                                port_spec.name in module.editable_input_ports)
                self.addTopLevelItem(item)
                self.port_spec_items[port_spec.name] = (port_spec, item)

            if self.port_type == 'input':
                for function in module.functions:
                    if not function.is_valid:
                        continue
                    port_spec, item = self.port_spec_items[function.name]
                    subitem = self.entry_klass(port_spec, function,
                                               self.types_visible)
                    self.function_map[function.real_id] = subitem
                    item.addChild(subitem)
                    subitem.setFirstColumnSpanned(True)
                    self.setItemWidget(subitem, 2, subitem.get_widget())
                    item.setExpanded(True)

                    # self.setItemWidget(item, 0, item.get_visible())
                    # self.setItemWidget(item, 1, item.get_connected())

                    # i = QTreeWidgetItem(self)
                    # self.addTopLevelItem(i)
                    # i.setText(2, port_spec.name)
                    # visible_checkbox = QtGui.QCheckBox()
                    # self.setItemWidget(i, 0, visible_checkbox)
                    # connceted_checkbox = QtGui.QCheckBox()
                    # connected_checkbox.setEnabled(False)
                    # self.setItemWidget(i, 1, connected_checkbox)

                # Highlight unset ports
                for _, item  in self.port_spec_items.itervalues():
                    item.calcUnset()

                self.sortItems(0, QtCore.Qt.AscendingOrder)

            # base_items = {}
            # # Create the base widget item for each descriptor
            # for descriptor in moduleHierarchy:
            #     baseName = descriptor.name
            #     base_package = descriptor.identifier
            #     baseItem = QMethodTreeWidgetItem(None,
            #                                      None,
            #                                      self,
            #                                      ([]
            #                                       <<  baseName
            #                                       << ''))
            #     base_items[descriptor] = baseItem

            # method_specs = {}
            # # do this in reverse to ensure proper overloading
            # # !!! NOTE: we have to use ***all*** input ports !!!
            # # because a subclass can overload a port with a 
            # # type that isn't a method
            # for descriptor in reversed(moduleHierarchy):
            #     method_specs.update((name, (descriptor, spec))
            #                         for name, spec in \
            #                             registry.module_ports('input', 
            #                                                   descriptor))

            # # add local registry last so that it takes precedence
            # method_specs.update((spec.name, (descriptor, spec))
            #                     for spec in module.port_spec_list
            #                     if spec.type == 'input')

            # for _, (desc, method_spec) in sorted(method_specs.iteritems()):
            #     if registry.is_method(method_spec):
            #         baseItem = base_items[desc]
            #         sig = method_spec.short_sigstring
            #         QMethodTreeWidgetItem(module,
            #                               method_spec,
            #                               baseItem,
            #                               ([]
            #                                << method_spec.name
            #                                << sig))

            # self.expandAll()
            # self.resizeColumnToContents(2) 
        # show invalid module attributes
        if module and not module.is_valid and self.port_type == 'input':
            for function in module.functions:
                if function.name in self.port_spec_items:
                    port_spec, item = self.port_spec_items[function.name]
                else:
                    sigstring = create_port_spec_string(
                        [(basic_identifier, "String")
                         for i in xrange(len(function.parameters))])
                    port_spec = PortSpec(name=function.name, type='input',
                                         sigstring=sigstring)
                    item = PortItem(port_spec,  False, False, False)
                self.addTopLevelItem(item)
                self.port_spec_items[port_spec.name] = (port_spec, item)
                subitem = self.entry_klass(port_spec, function)
                self.function_map[function.real_id] = subitem
                item.addChild(subitem)
                subitem.setFirstColumnSpanned(True)
                self.setItemWidget(subitem, 2, subitem.get_widget())
                item.setExpanded(True)

    def item_clicked(self, item, col):
        if item.parent() is not None:
            return

        if self.port_type == 'input':
            visible_ports = self.module.visible_input_ports
            editable_ports = self.module.editable_input_ports
        elif self.port_type == 'output':
            visible_ports = self.module.visible_output_ports
        else:
            raise TypeError("Unknown port type: '%s'" % self.port_type)

        if col == 0:
            if item.is_constant() and len(item.port_spec.port_spec_items)>0:
                item.set_editable(not item.is_editable)
                if item.is_editable:
                    editable_ports.add(item.port_spec.name)
                else:
                    editable_ports.discard(item.port_spec.name)
                self.controller.flush_delayed_actions()
                self.controller.add_annotation((self.module.INLINE_WIDGET_ANNOTATION,
                                                ','.join(editable_ports)),
                                               self.module.id)
                self.controller.current_pipeline_scene.recreate_module(
                    self.controller.current_pipeline, self.module.id)
        if col == 1:
            if item.is_optional:
                item.set_visible(not item.is_visible)
                if item.is_visible:
                    visible_ports.add(item.port_spec.name)
                else:
                    visible_ports.discard(item.port_spec.name)
                self.controller.flush_delayed_actions()
                self.controller.current_pipeline_scene.recreate_module(
                    self.controller.current_pipeline, self.module.id)
        if col == 3:
            if item.isExpanded():
                item.setExpanded(False)
            elif item.childCount() > 0:
                item.setExpanded(True)
            elif item.childCount() == 0 and item.is_constant():
                self.do_add_method(item.port_spec, item)
        
    def set_controller(self, controller):
        self.controller = controller

    def update_method(self, subitem, port_name, widgets, labels, real_id=-1):
        #print 'updateMethod called', port_name
        if self.controller:
            _, item = self.port_spec_items[port_name]
            str_values = []
            query_methods = []
            for w in widgets:
                str_values.append(str(w.contents()))
                if hasattr(w, 'query_method'):
                    query_methods.append(w.query_method())
            if real_id < 0:
                should_replace = False
            else:
                should_replace = True
            self.controller.update_function(self.module,
                                            port_name,
                                            str_values,
                                            real_id,
                                            [str(label.alias)
                                             for label in labels],
                                            query_methods,
                                            should_replace)

            # FIXME need to get the function set on the item somehow
            # HACK for now
            for function in self.module.functions:
                if function.real_id not in self.function_map:
                    self.function_map[function.real_id] = subitem
                    subitem.function = function

            # make the scene display the fact that we have a parameter
            # by dimming the port
            # self.controller.flush_delayed_actions()
            self.controller.current_pipeline_scene.update_module_functions(
                self.controller.current_pipeline, self.module.id)
                                            
    def delete_method(self, subitem, port_name, real_id=None):
        _, item = self.port_spec_items[port_name]
        item.removeChild(subitem)

        if real_id is not None and self.controller:
            #print "got to delete"
            self.controller.delete_function(real_id, self.module.id)

            # make the scene display the fact that we have lost the
            # parameter by undimming the port
            # self.controller.flush_delayed_actions()
            self.controller.current_pipeline_scene.update_module_functions(
                self.controller.current_pipeline, self.module.id)

        # how to delete items...x
        # subitem.deleteLater()
            
    def do_add_method(self, port_spec, item):
        """do_add_method(port_spec: PortSpec,
                         item:      PortItem) -> None

        Displays a new method for the port.
        """

        subitem = self.entry_klass(port_spec)
        item.addChild(subitem)
        subitem.setFirstColumnSpanned(True)
        self.setItemWidget(subitem, 2, subitem.get_widget())
        item.setExpanded(True)
        if len(port_spec.descriptors()) == 0:
            self.update_method(subitem, port_spec.name, [], [])

    def add_method(self, port_name):
        port_spec, item = self.port_spec_items[port_name]
        self.do_add_method(port_spec, item)
        
    def contextMenuEvent(self, event):
        # Just dispatches the menu event to the widget item
        item = self.itemAt(event.pos())
        if item:
            item.contextMenuEvent(event, self)

class QPortsPane(QtGui.QWidget, QToolWindowInterface):
    def __init__(self, port_type, parent=None, flags=QtCore.Qt.Widget):
        QtGui.QWidget.__init__(self, parent, flags)
        self.port_type = port_type
        self.build_widget()
        self.controller = None

    def build_widget(self):
        self.tree_widget = PortsList(self.port_type)
        layout = QtGui.QHBoxLayout()
        layout.setMargin(0)
        layout.addWidget(self.tree_widget)
        self.setLayout(layout)
        self.setWindowTitle('%s Ports' % self.port_type.capitalize())

    def set_controller(self, controller):
        self.controller = controller
        self.tree_widget.set_controller(controller)

    def update_module(self, module):
        self.tree_widget.update_module(module)
    



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from auto_gen import DBLog as _DBLog
from auto_gen import DBAbstraction, DBModule, DBGroup
from id_scope import IdScope

import copy

class DBLog(_DBLog):

    def __init__(self, *args, **kwargs):
        _DBLog.__init__(self, *args, **kwargs)
        self.id_scope = IdScope(1,
                              {DBAbstraction.vtType: DBModule.vtType,
                               DBGroup.vtType: DBModule.vtType})

    def __copy__(self):
        return DBLog.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = _DBLog.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = DBLog
        cp.id_scope = copy.copy(self.id_scope)
        
        return cp

    @staticmethod
    def update_version(old_obj, trans_dict, new_obj=None):
        if new_obj is None:
            new_obj = DBLog()
        new_obj = _DBLog.update_version(old_obj, trans_dict, new_obj)
        new_obj.update_id_scope()
        return new_obj

    def update_id_scope(self):
        pass



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import copy
import os.path
from vistrails.core.system import current_user, current_time
from vistrails.core.mashup.alias import Alias
from vistrails.core.mashup.component import Component
from vistrails.core.mashup.mashup import Mashup

class MashupController(object):
    def __init__(self, originalController, vt_controller, vt_version, mshptrail=None):
        self.vtController = vt_controller
        self.originalController = originalController
        self.vtVersion = vt_version
        self.vtPipeline = self.vtController.vistrail.getPipeline(self.vtVersion)
        self.vtPipeline.validate()
        self.mshptrail = mshptrail
        self.id_scope = mshptrail.id_scope
        self.currentVersion = -1
        self.currentMashup = None
        self._changed = False

    def setChanged(self, on):
        self._changed = on
        self.originalController.set_changed(True)
        
    def setCurrentVersion(self, version):
        self.currentVersion = version
        self.vtPipeline = self.vtController.vistrail.getPipeline(self.vtVersion)
        if version > -1:
            self.currentMashup = self.mshptrail.getMashup(version)
            self.updatePipelineAliasesFromCurrentMashup()
            
    def getVistrailParam(self, alias):
        if self.vtPipeline:
            return self.vtPipeline.db_get_object(alias.component.vttype,
                                                  alias.component.vtid)
        return None
    
    def execute(self, params):
        if self.vtPipeline and self.vtController:
            mashup_id = self.mshptrail.id
            mashup_version = self.currentVersion
            reason = "mashup::%s::%s"%(str(mashup_id), mashup_version)
            result = self.vtController.execute_current_workflow(custom_params=params,
                                                              reason=reason)
            self.originalController.set_changed(True)
            return result
        return ([], False)
            
    def updateCurrentTag(self, name):
        if self.mshptrail.changeTag(self.currentVersion, name, current_user(),
                                 current_time()):
            self.setChanged(True)
            return True
        else:
            return False
    
    def moveTag(self, from_version, to_version, name):
        tag = self.mshptrail.getTagForActionId(from_version)
        if tag:
            self.mshptrail.removeTagByActionId(from_version)
            self.mshptrail.addTag(to_version, tag, user=current_user(),
                                  date=current_time())
            
    def getCurrentTag(self):
        return self.mshptrail.getTagForActionId(self.currentVersion)
    
    def versionHasTag(self, version):
        return self.mshptrail.hasTagForActionId(version)
    
    def hasTagWithName(self, name):
        return self.mshptrail.hasTagWithName(name)
    
    def getVistrailName(self):
        name = ''
        locator = self.currentMashup.vtid
        if locator is not None:
            if locator.name is None:
                name = ''
            else:
                name = os.path.split(locator.name)[1]
            if name == '':
                name = self.controller.vtController.name
        return name
        
    def resetVistrailPipeline(self):
        self.vtController.change_selected_version(self.vtVersion)
    
    def getVistrailWorkflowTag(self):
        return self.vtController.get_pipeline_name(self.vtVersion)
    
    def reorderAliases(self, new_order):
        if self.currentMashup:
            new_aliases = []
            pos = 0
            for old_pos in new_order:
                alias = self.currentMashup.alias_list[old_pos].do_copy(
                               new_ids=True, id_scope=self.mshptrail.id_scope,
                               id_remap={})
                alias.component.pos = pos
                new_aliases.append(alias)
                pos += 1
            return self.createMashupVersion(new_aliases, quiet=False)
                
    def updateAlias(self, alias):
        """updateAlias(alias)-> long
        This will create a version with an alias change (can't be a position
        change). Position changes are taken care in reorderAliases method. 
        
        """
        #print " controller updateAlias ", alias
        new_aliases = []

        if self.currentMashup:
            for a in self.currentMashup.alias_list:
                if a.id != alias.id:
                    calias = a.do_copy(new_ids=True,
                                       id_scope=self.mshptrail.id_scope,
                                       id_remap={})
                else:
                    #print "found alias: ", a
                    calias = alias.do_copy(new_ids=True,
                                           id_scope=self.mshptrail.id_scope,
                                           id_remap={})
                new_aliases.append(calias)
        return self.createMashupVersion(new_aliases, quiet=False)
        
    def updateAliasFromParam(self, param):
        add_alias = True
        new_aliases = []
        pos = 0
        for alias in self.currentMashup.alias_list:
            if alias.component.vtid != param.id:
                calias = alias.do_copy(new_ids=True,
                                       id_scope=self.mshptrail.id_scope,
                                       id_remap={})
                calias.component.pos = pos
                new_aliases.append(calias)
                pos += 1
            else:
                #print "found alias: ", alias
                add_alias = False
                if param.alias != '':
                    new_alias = alias.do_copy(new_ids=True,
                                              id_scope=self.mshptrail.id_scope,
                                              id_remap={})
                    new_alias.name = param.alias
                    new_aliases.append(new_alias)
                    pos += 1
        if add_alias:
            parameter = self.vtPipeline.db_get_object(param.dbtype, param.id)
            cid = self.id_scope.getNewId('mashup_component')
            aid = self.id_scope.getNewId('mashup_alias')
            component = Component(cid, parameter.vtType, 
                                  parameter.real_id, param.parent_dbtype, 
                                  param.parent_id,
                                  param.mId, parameter.type, 
                                  parameter.strValue, parameter.pos, 
                                  pos, "")
            alias = Alias(aid, param.alias, component) 
            new_aliases.append(alias)
            self.vtPipeline.add_alias(param.alias, param.type, param.id,
                                      param.parent_dbtype, param.parent_id,
                                      param.mId)
        else:
            self.vtPipeline.change_alias(param.alias, param.type, param.id,
                                         param.parent_dbtype, param.parent_id,
                                         param.mId)
        
        return self.createMashupVersion(new_aliases, quiet=False)
        
    def updateAliasesFromPipeline(self, pipeline):
        """updateAliasesFromPipeline(self, pipeline) -> long
        This will generate a new mashup by updating the aliases of the current 
        mashup according to the aliases in a pipeline. This assumes that the 
        mashup's current aliases are different from pipeline aliases by at most 
        one change (eg., an alias rename, an alias addition, an alias removal)
        
        """
        pip_aliases = pipeline.aliases.keys()
        mashup_aliases = [a.name for a in self.currentMashup.alias_list]
        new_aliases = []
        if len(pip_aliases) == len(mashup_aliases):
            #an alias probably changed its name or its value    
            old_a = None
            new_a = None
            for a in self.currentMashup.alias_list:
                if a.name not in pip_aliases:
                    old_a = a.do_copy(new_ids=True,
                                      id_scope=self.mshptrail.id_scope,
                                      id_remap={})
                    new_aliases.append(old_a)
                else:
                    new_aliases.append(a)
            for a in pip_aliases:
                if a not in mashup_aliases:
                    new_a = (a, pipeline.aliases[a])
            if old_a is not None and new_a is not None:
                (a, info) = new_a
                parameter = pipeline.db_get_object(info[0],info[1])
                old_a.name = a
                old_a.component.vttype = parameter.vtType 
                old_a.component.vtid = parameter.real_id
                old_a.component.vtparent_type = info[2]
                old_a.component.vt_parent_id = info[3]
                old_a.component.mid = info[4]
                old_a.component.type = parameter.type
                old_a.component.val = parameter.strValue
                old_a.component.vtpos = parameter.pos
            
        elif len(pip_aliases) < len(mashup_aliases):
            # an alias was removed
            pos = 0
            for a in self.currentMashup.alias_list:
                if a.name in pip_aliases:
                    alias = a.do_copy(new_ids=True,
                                      id_scope=self.mshptrail.id_scope,
                                      id_remap={})
                    alias.component.pos = pos
                    new_aliases.append(alias)
                    pos += 1
        else:
            #an alias was added
            pos = len(mashup_aliases)
            new_aliases = [a for a in self.currentMashup.alias_list]
            for a in pip_aliases:
                if a not in mashup_aliases:
                    info = pipeline.aliases[a]
                    parameter = pipeline.db_get_object(info[0],info[1])
                    cid = self.id_scope.getNewId('mashup_component')
                    aid = self.id_scope.getNewId('mashup_alias')
                    component = Component(cid, parameter.vtType, 
                                          parameter.real_id, info[2], info[3],
                                          info[4], parameter.type, 
                                          parameter.strValue, parameter.pos, 
                                          pos, "")
                    alias = Alias(aid, a, component) 
                    new_aliases.append(alias)
                    pos += 1
        
        return self.createMashupVersion(new_aliases, quiet=False)
        
    def updatePipelineAliasesFromCurrentMashup(self):
        self.resetVistrailPipeline()
        self.vtPipeline = copy.copy(self.vtController.current_pipeline)
        #first we clear all aliases in pipeline
        to_remove = self.vtPipeline.aliases.values()
        for (type, oId, parentType, parentId, mid) in to_remove:
            self.vtPipeline.remove_alias(type, oId, parentType, parentId, mid)
            parameter = self.vtPipeline.db_get_object(type,oId)
            parameter.alias = ''
            
        #now we populate the pipeline according to the aliases in the mashup 
        for alias in self.currentMashup.alias_list:
            self.vtPipeline.add_alias(alias.name, alias.component.vttype, 
                                      alias.component.vtid,
                                      alias.component.vtparent_type, 
                                      alias.component.vtparent_id,
                                      alias.component.vtmid)
            parameter = self.vtPipeline.db_get_object(alias.component.vttype,
                                                      alias.component.vtid)
            parameter.alias = alias.name
        
    def getMashupName(self, version=-1):
        action_map = self.mshptrail.actionMap
        if version == -1:
            version = self.currentVersion
        count = 0
        while True:
            hasTag = self.mshptrail.hasTagForActionId(version)
            if hasTag or version <= 1:
                if hasTag:
                    name = self.mshptrail.getTagForActionId(version)
                else:
                    name = "ROOT"
                count_str = ""
                if count > 0:
                    count_str = " + " + str(count)
                return name + count_str
            version = action_map[version].parent_id
            count += 1
            
    def findFirstTaggedParent(self, version):
        action_map = self.mshptrail.actionMap
        version = action_map[version].parent_id
        while True:
            hasTag = self.mshptrail.hasTagForActionId(version)
            if hasTag or version <= 1:
                name = ""
                if hasTag:
                    name = self.mshptrail.getTagForActionId(version)
                return (version, name)
            version = action_map[version].parent_id
            
    def removeAlias(self, name):
        """removeAlias(name: str) -> long
        This will create a new version of the mashup without alias name, add it
        to the trail and set the version as the current version. It will return 
        the version number 
        """
        new_aliases = []
        if self.currentMashup:
            pos = 0
            for alias in self.currentMashup.alias_list:
                if alias.name != name:
                    calias = alias.do_copy(new_ids=True,
                                           id_scope=self.mshptrail.id_scope,
                                           id_remap={})
                    calias.component.pos = pos
                    new_aliases.append(calias)
                    pos += 1
            return self.createMashupVersion(alias_list=new_aliases, quiet=False)
            
    def createMashupVersion(self, alias_list, quiet=False):
        id = self.id_scope.getNewId('mashup')
        mashup = Mashup(id=id, name="mashup%s"%id, 
                        vtid=self.currentMashup.vtid, 
                        version=self.currentMashup.version, 
                        alias_list=alias_list)
        currVersion = self.mshptrail.addVersion(parent_id=self.currentVersion,
                                                mashup=mashup, 
                                                user=current_user(),
                                                date=current_time())
        self.mshptrail.currentVersion = currVersion
        self.currentMashup = mashup
        #print "created new mashup ", currVersion
        self.setCurrentVersion(currVersion, quiet)
        self.setChanged(True)
        return currVersion



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.db.versions.v1_0_4.domain import DBVistrail, DBVistrailVariable, \
                                      DBWorkflow, DBLog, DBRegistry, \
                                      DBAdd, DBChange, DBDelete, \
                                      DBPortSpec, DBPortSpecItem, \
                                      DBParameterExploration, \
                                      DBPEParameter, DBPEFunction, \
                                      IdScope, DBAbstraction, \
                                      DBModule, DBGroup, DBAnnotation, \
                                      DBActionAnnotation, DBStartup, \
                                      DBConfigKey, DBConfigBool, DBConfigStr, \
                                      DBConfigInt, DBConfigFloat, \
                                      DBConfiguration, DBStartupPackage, \
                                      DBLoopIteration, DBLoopExec, \
                                      DBModuleExec, DBGroupExec

from vistrails.db.services.vistrail import materializeWorkflow
from xml.dom.minidom import parseString
from itertools import izip
import os
import shutil
import string
import tempfile
from ast import literal_eval
import unittest

id_scope = None

def translateVistrail(_vistrail):
    """ Translate old annotation based vistrail variables to new
        DBVistrailVariable class """
    global id_scope

    def update_workflow(old_obj, trans_dict):
        return DBWorkflow.update_version(old_obj.db_workflow, 
                                         trans_dict, DBWorkflow())

    translate_dict = {'DBGroup': {'workflow': update_workflow}}
    vistrail = DBVistrail()
    id_scope = vistrail.idScope
    vistrail = DBVistrail.update_version(_vistrail, translate_dict, vistrail)

    vistrail.db_version = '1.0.4'
    return vistrail

def translateWorkflow(_workflow):
    global id_scope
    def update_workflow(old_obj, translate_dict):
        return DBWorkflow.update_version(old_obj.db_workflow, translate_dict)
    translate_dict = {'DBGroup': {'workflow': update_workflow}}

    workflow = DBWorkflow()
    id_scope = IdScope(remap={DBAbstraction.vtType: DBModule.vtType, DBGroup.vtType: DBModule.vtType})
    workflow = DBWorkflow.update_version(_workflow, translate_dict, workflow)
    workflow.db_version = '1.0.4'
    return workflow

def translateLog(_log):
    id_scope = _log.id_scope
    def update_loop_execs(old_obj, translate_dict):
        if len(old_obj.db_loop_execs) == 0:
            return []
        # create new LoopExec and add existing as LoopIteration's
        ts_start = None
        ts_end = None
        new_iters = []
        for loop_exec in old_obj.db_loop_execs:
            item_execs = []
            for item_exec in loop_exec.db_item_execs:
                if item_exec.vtType == DBModuleExec.vtType:
                    item_execs.append(DBModuleExec.update_version(item_exec,
                                                                  translate_dict))
                if item_exec.vtType == DBGroupExec.vtType:
                    item_execs.append(DBGroupExec.update_version(item_exec,
                                                                 translate_dict))
                # skip DBLoopExecs
            new_iter = DBLoopIteration(item_execs=item_execs,
                                       id=id_scope.getNewId(DBLoopIteration.vtType),
                                       ts_start=loop_exec.db_ts_start,
                                       ts_end=loop_exec.db_ts_end,
                                       iteration=loop_exec.db_iteration,
                                       completed=loop_exec.db_completed,
                                       error=loop_exec.db_error)
            new_iters.append(new_iter)
            if ts_start is None or ts_start > loop_exec.db_ts_start:
                ts_start = loop_exec.db_ts_start
            if ts_end is None or ts_end < loop_exec.db_ts_end:
                ts_end = loop_exec.db_ts_end
        return [DBLoopExec(id=id_scope.getNewId(DBLoopExec.vtType),
                                       ts_start=ts_start,
                                       ts_end=ts_end,
                                       loop_iterations=new_iters)]

    def update_item_execs(old_obj, translate_dict):
        # remove loop_execs and add them as loop_iterations under a single loop_exec
        new_objs = []
        ts_start = None
        ts_end = None
        new_iters = []
        for obj in old_obj.db_item_execs:
            if obj.vtType == DBModuleExec.vtType:
                new_objs.append(DBModuleExec.update_version(obj,
                                                            translate_dict))
            if obj.vtType == DBGroupExec.vtType:
                new_objs.append(DBGroupExec.update_version(obj,
                                                           translate_dict))
            if obj.vtType == DBLoopExec.vtType:
                item_execs = []
                for item_exec in obj.db_item_execs:
                    if item_exec.vtType == DBModuleExec.vtType:
                        item_execs.append(DBModuleExec.update_version(item_exec,
                                                                      translate_dict))
                    if item_exec.vtType == DBGroupExec.vtType:
                        item_execs.append(DBGroupExec.update_version(item_exec,
                                                                     translate_dict))
                    # skip DBLoopExecs
                new_iter = DBLoopIteration(item_execs=item_execs,
                                           id=id_scope.getNewId(DBLoopIteration.vtType),
                                           ts_start=obj.db_ts_start,
                                           ts_end=obj.db_ts_end,
                                           iteration=obj.db_iteration,
                                           completed=obj.db_completed,
                                           error=obj.db_error)
                new_iters.append(new_iter)
                if ts_start is None or ts_start > obj.db_ts_start:
                    ts_start = obj.db_ts_start
                if ts_end is None or ts_end < obj.db_ts_end:
                    ts_end = obj.db_ts_end
        if len(new_iters):
            new_objs.append(DBLoopExec(id=id_scope.getNewId(DBLoopExec.vtType),
                                       ts_start=ts_start,
                                       ts_end=ts_end,
                                       loop_iterations=new_iters))
        return new_objs

        
        return DBWorkflow.update_version(old_obj.db_workflow, translate_dict)
    translate_dict = {'DBModuleExec': {'loop_execs': update_loop_execs},
                      'DBGroupExec': {'item_execs': update_item_execs}}
    log = DBLog.update_version(_log, translate_dict)
    log.db_version = '1.0.4'
    return log

def translateRegistry(_registry):
    global id_scope
    translate_dict = {}
    registry = DBRegistry()
    id_scope = registry.idScope
    registry = DBRegistry.update_version(_registry, translate_dict, registry)
    registry.db_version = '1.0.4'
    return registry

def translateStartup(_startup):
    # format is {<old_name>: <new_name>} or
    # {<old_name>: (<new_name> | None, [conversion_f | None, inner_d | None])
    # conversion_f is a function that mutates the value and
    # inner_d recurses the translation for inner configurations

    dot_vt_path = None
    startup_dir = os.path.dirname(_startup._filename)
    if os.path.isabs(startup_dir):
        dot_vt_path = startup_dir

    def change_value(cls, conv_f, _key, t_dict):
        cls_name = cls.__name__
        old_t_value = None
        if cls_name in t_dict:
            if 'value' in t_dict[cls_name]:
                old_t_value = t_dict[cls_name]['value']
        else:
            t_dict[cls_name] = {}
        t_dict[cls_name]['value'] = conv_f
        new_value = cls.update_version(_key.db_value, t_dict)
        del t_dict[cls_name]['value']
        if old_t_value is not None:
            t_dict[cls_name]['value'] = old_t_value
        if len(t_dict[cls_name]) < 1:
            del t_dict[cls_name]
        return new_value

    def invert_bool(_key, t_dict):
        def invert_value(_value, t):
            return unicode(not (_value.db_value.lower() == 'true'))
        return change_value(DBConfigBool, invert_value, _key, t_dict)

    def use_dirname(_key, t_dict):
        def get_dirname(_value, t):
            dir_name = os.path.dirname(_value.db_value)
            if os.path.isabs(dir_name) and dir_name.startswith(dot_vt_path):
                dir_name = dir_name[len(dot_vt_path)+1:]
                if not dir_name.strip():
                    dir_name = "logs"
            return dir_name
        return change_value(DBConfigStr, get_dirname, _key, t_dict)

    def update_dirname(_key, t_dict):
        def change_dirname(_value, t):
            abs_path = _value.db_value
            if os.path.isabs(abs_path) and abs_path.startswith(dot_vt_path):
                return abs_path[len(dot_vt_path)+1:]
            return abs_path
        return change_value(DBConfigStr, change_dirname, _key, t_dict)

    def pdf_bool_to_str(_value):
        s_val = "PNG"
        if _value.db_value.lower() == "true":
            s_val = "PDF"
        return DBConfigStr(value=s_val)

    t = {'alwaysShowDebugPopup': 'showDebugPopups',
         'autosave': 'autoSave',
         'errorOnConnectionTypeerror': 'showConnectionErrors',
         'errorOnVariantTypeerror': 'showVariantErrors',
         'interactiveMode': ('batch', invert_bool),
         'logFile': ('logDir', use_dirname),
         'logger': None, # DELETE
         'maxMemory': None,
         'minMemory': None,
         'nologger': ('executionLog', invert_bool),
         'pythonPrompt': None,
         'reviewMode': None, # unknown what this was/is
         'shell': ('shell', None, {'font_face': 'fontFace',
                                   'font_size': 'fontSize'}),
         'showMovies': None,
         'showSpreadsheetOnly': ('showWindow', invert_bool),
         'spreadsheetDumpCells': 'outputDirectory',
         'upgradeOn': 'upgrades',
         'useCache': 'cache',
         'verbosenessLevel': 'debugLevel',
         'webRepositoryLogin': 'webRespositoryUser',
         'evolutionGraph': 'withVersionTree',
         'workflowGraph': 'withWorkflow',
         # 'workflowInfo': 'outputWorkflowInfo',
         'workflowInfo': 'outputDirectory',
         'executeWorkflows': 'execute',

         'thumbs': ('thumbs', None,
                    {'cacheDirectory': ('cacheDir', update_dirname)}),
         'abstractionsDirectory': ('subworkflowsDir', update_dirname),
         'userPackageDirectory': ('userPackageDir', update_dirname),
         'temporaryDirectory': 'temporaryDir',
         'dataDirectory': 'dataDir',
         'fileDirectory': 'fileDir',
         'packageDirectory': 'packageDir',

         }

    # these are moving to package configurations
    t_package = {
         'spreadsheetDumpPDF': ('spreadsheet', 'dumpfileType', pdf_bool_to_str),
         'fixedSpreadsheetCells': ('spreadsheet', 'fixedCellSize'),
    }

    def get_key_name_update(new_name):
        def update_key_name(_key, t_dict):
            return new_name
        return update_key_name

    def update_config_keys(_config, t_dict):
        # just update __startup_translate__ as we recurse,
        # update_key_name and update_key_value take care of the rest
        cur_t = t_dict['__startup_translate__']
        cur_t_pkg = t_dict['__startup_translate_pkg__']
        new_keys = []
        for key in _config.db_config_keys:
            to_delete = False
            new_name = None
            conv_f = None
            inner_t = None
            if key.db_name in cur_t:
                t_obj = cur_t[key.db_name]
                if type(t_obj) == tuple:
                    new_name = t_obj[0]
                    if len(t_obj) > 1:
                        conv_f = t_obj[1]
                    if len(t_obj) > 2:
                        inner_t = t_obj[2]
                elif t_obj is None:
                    to_delete = True
                else:
                    new_name = t_obj

            if key.db_name in cur_t_pkg:
                updated_key = DBConfigKey.update_version(key, t_dict)
                pkg_t = cur_t_pkg[key.db_name]
                pkg_name = pkg_t[0]
                key_name = pkg_t[1]
                config_dict = t_dict['__startup_package_config__']

                if pkg_name not in config_dict:
                    config_dict[pkg_name] = []
                if len(pkg_t) > 2:
                    value_f = pkg_t[2]
                    new_value = value_f(key.db_value)
                else:
                    # have to update version here?
                    new_value = updated_key.db_value
                config_dict[pkg_name].append(DBConfigKey(name=key_name,
                                                         value=new_value))

            # if we have already set a key, don't override the new
            # setting in general (may want to if swapping existing
            # names, but this should be the exception)
            elif (not to_delete and
                (new_name == key.db_name or
                 new_name not in _config.db_config_keys_name_index)):
                # always overwrite DBConfigKey settings so recursion
                # doesn't wrap over itself
                key_t_dict = {}
                if new_name is not None and new_name != key.db_name:
                    key_t_dict['name'] = get_key_name_update(new_name)
                if conv_f is not None:
                    key_t_dict['value'] = conv_f
                old_t_name = None
                old_t_value = None
                if len(key_t_dict) > 0:
                    if 'DBConfigKey' in t_dict:
                        if 'name' in t_dict['DBConfigKey']:
                            old_t_name = t_dict['DBConfigKey']['name']
                            del t_dict['DBConfigKey']['name']
                        if 'value' in t_dict['DBConfigKey']:
                            old_t_value = t_dict['DBConfigKey']['value']
                            del t_dict['DBConfigKey']['value']
                        t_dict['DBConfigKey'].update(key_t_dict)
                    else:
                        t_dict['DBConfigKey'] = key_t_dict
                if inner_t is not None:
                    t_dict['__startup_translate__'] = inner_t
                new_keys.append(DBConfigKey.update_version(key, t_dict))
                if inner_t is not None:
                    t_dict['__startup_translate__'] = cur_t
                if len(key_t_dict) > 0:
                    if 'name' in t_dict['DBConfigKey']:
                        del t_dict['DBConfigKey']['name']
                    if 'value' in t_dict['DBConfigKey']:
                        del t_dict['DBConfigKey']['value']
                    if old_t_name is not None:
                        t_dict['DBConfigKey']['name'] = old_t_name
                    if old_t_value is not None:
                        t_dict['DBConfigKey']['value'] = old_t_value
                    if len(t_dict['DBConfigKey']) < 1:
                        del t_dict['DBConfigKey']
        return new_keys

    def update_package_config(_startup_pkg, t_dict):
        _config = _startup_pkg.db_configuration
        package_config = t_dict['__startup_package_config__']
        if _config is not None:
            config = DBConfiguration.update_version(_config, t_dict)
            if _startup_pkg.db_name in package_config:
                for new_key in package_config[_startup_pkg.db_name]:
                    if new_key.db_name in config.db_config_keys_name_index:
                        # already have, don't replace new setting
                        continue
                    else:
                        config.db_add_config_key(new_key)
                del package_config[_startup_pkg.db_name]
            return config
        return None

    # need to recurse both key name changes and key value changes
    # (could also change types...)
    translate_dict = {'DBConfiguration': {'config_keys': update_config_keys},
                      'DBStartupPackage': {'configuration':
                                           update_package_config},
                      '__startup_translate__': t,
                      '__startup_translate_pkg__': t_package,
                      '__startup_package_config__': dict()}
    startup = DBStartup()
    startup = DBStartup.update_version(_startup, translate_dict, startup)

    # add any translations to packages, creating startup_pkg and
    # configuration if necessary
    for pkg, keys in translate_dict['__startup_package_config__'].iteritems():
        if pkg in startup.db_enabled_packages.db_packages_name_index:
            s_pkg = startup.db_enabled_packages.db_packages_name_index[pkg]
        elif pkg in startup.db_disabled_packages.db_packages_name_index:
            s_pkg = startup.db_disabled_packages.db_packages_name_index[pkg]
        else:
            s_pkg = DBStartupPackage(name=pkg)
            startup.db_disabled_packages.db_add_package(s_pkg)
        if s_pkg.db_configuration is None:
            s_pkg.db_configuration = DBConfiguration(config_keys=keys)
        else:
            for new_key in keys:
                if new_key.db_name not in s_pkg.db_configuration:
                    # only add if new setting doesn't already exist
                    s_pkg.db_configuration.db_add_config_key(new_key)

    startup.db_version = '1.0.4'
    return startup

class TestTranslate(unittest.TestCase):
    def test_startup_update(self):
        from vistrails.db.services.io import open_startup_from_xml
        from vistrails.core.system import vistrails_root_directory
        import os

        startup_tmpl = os.path.join(vistrails_root_directory(),
                                    'tests', 'resources',
                                    'startup-0.1.xml.tmpl')
        f = open(startup_tmpl, 'r')
        template = string.Template(f.read())

        startup_dir = tempfile.mkdtemp(prefix="vt_startup")
        startup_fname = os.path.join(startup_dir, "startup.xml")
        with open(startup_fname, 'w') as f:
            f.write(template.substitute({'startup_dir': startup_dir}))
        try:
            # FIXME need to generate startup from local path
            startup = open_startup_from_xml(startup_fname)
            name_idx = startup.db_configuration.db_config_keys_name_index
            self.assertNotIn('nologger', name_idx)
            self.assertIn('executionLog', name_idx)
            self.assertFalse(
                name_idx['executionLog'].db_value.db_value.lower() == 'true')
            self.assertNotIn('showMovies', name_idx)
            self.assertIn('logDir', name_idx)
            self.assertEqual(name_idx['logDir'].db_value.db_value, 'logs')
            self.assertIn('userPackageDir', name_idx)
            self.assertEqual(name_idx['userPackageDir'].db_value.db_value,
                             'userpackages')
            self.assertIn('thumbs', name_idx)
            thumbs_name_idx = \
                    name_idx['thumbs'].db_value.db_config_keys_name_index
            self.assertIn('cacheDir', thumbs_name_idx)
            self.assertEqual(thumbs_name_idx['cacheDir'].db_value.db_value,
                             '/path/to/thumbs')
            self.assertIn('subworkflowsDir', name_idx)
            self.assertEqual(name_idx['subworkflowsDir'].db_value.db_value,
                             'subworkflows')

            # note: have checked with spreadsheet removed from all
            # packages list, too
            # TODO: make this a permanent test (new template?)
            self.assertNotIn('fixedSpreadsheetCells', name_idx)
            enabled_names = startup.db_enabled_packages.db_packages_name_index
            self.assertIn('spreadsheet', enabled_names)
            spreadsheet_config = enabled_names['spreadsheet'].db_configuration
            self.assertIsNotNone(spreadsheet_config)
            spreadsheet_name_idx = spreadsheet_config.db_config_keys_name_index
            self.assertIn('fixedCellSize', spreadsheet_name_idx)
            self.assertTrue(spreadsheet_name_idx['fixedCellSize'].db_value.db_value.lower() == "true")
            self.assertIn('dumpfileType', spreadsheet_name_idx)
            self.assertEqual(spreadsheet_name_idx['dumpfileType'].db_value.db_value, "PNG")
        finally:
            shutil.rmtree(startup_dir)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from PyQt4 import QtCore, QtGui

class BaseView(object):
    """ BaseView is the base class for the views in VisTrails.


    """

    def __init__(self):
        self.controller = None
        self.title = None
        self.index = -1
        self.tab_idx = -1

        self.palette_layout = {}
        self.set_default_layout()
        self.action_links = {}
        self.action_defaults = {}
        BaseView.set_action_defaults(self)
        self.set_action_defaults()
        self.set_action_links()
        self.detachable = False
        self.vistrail_view = None

    def set_default_layout(self):
        raise NotImplementedError("Class must define the layout as self.palette_layout")

    def set_palette_layout(self, layout):
        self.palette_layout = layout

    def set_action_links(self):
        raise NotImplementedError("Class must define the action links")

    def set_action_defaults(self):
        self.action_defaults.update(
            {'history'    : [('setEnabled', False, True)],
             'search'     : [('setEnabled', False, True)],
             'explore'    : [('setEnabled', False, True)],
             'provenance' : [('setEnabled', False, True)],
             'mashup'     : [('setEnabled', False, True)]
             })

    def set_title(self, title):
        self.title = title

    def get_title(self):
        return self.title

    def set_long_title(self, title):
        self.long_title = title
        
    def get_long_title(self):
        return self.title
    
    def get_index(self):
        return self.index

    def set_index(self, index):
        self.index = index

    def get_tab_idx(self):
        return self.tab_idx

    def set_tab_idx(self, tab_idx):
        self.tab_idx = tab_idx

    def set_controller(self, controller):
        pass
    
    def set_vistrail_view(self, view):
        self.vistrail_view = view
        
    def get_vistrail_view(self):
        return self.vistrail_view
    
    def changeEvent(self, event):
        if event.type() == QtCore.QEvent.WindowTitleChange:
            self.emit(QtCore.SIGNAL("windowTitleChanged"), self)
        QtGui.QWidget.changeEvent(self, event)



#!/usr/bin/env python

# This example demonstrates the use of 2D text.

import vtk

# Create a sphere source, mapper, and actor
sphere = vtk.vtkSphereSource()

sphereMapper = vtk.vtkPolyDataMapper()
sphereMapper.SetInputConnection(sphere.GetOutputPort())
sphereMapper.GlobalImmediateModeRenderingOn()
sphereActor = vtk.vtkLODActor()
sphereActor.SetMapper(sphereMapper)

# Create a scaled text actor. 
# Set the text, font, justification, and properties (bold, italics,
# etc.).
textActor = vtk.vtkTextActor()
textActor.ScaledTextOn()
textActor.SetDisplayPosition(90, 50)
textActor.SetInput("This is a sphere")

# Set coordinates to match the old vtkScaledTextActor default value
textActor.GetPosition2Coordinate().SetCoordinateSystemToNormalizedViewport()
textActor.GetPosition2Coordinate().SetValue(0.6, 0.1)

tprop = textActor.GetTextProperty()
tprop.SetFontSize(18)
tprop.SetFontFamilyToArial()
tprop.SetJustificationToCentered()
tprop.BoldOn()
tprop.ItalicOn()
tprop.ShadowOn()
tprop.SetColor(0, 0, 1)

# Create the Renderer, RenderWindow, RenderWindowInteractor
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer; set the background and size; zoom
# in; and render.
ren.AddActor2D(textActor)
ren.AddActor(sphereActor)

ren.SetBackground(1, 1, 1)
renWin.SetSize(250, 125)
ren.ResetCamera()
ren.GetActiveCamera().Zoom(1.5)

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""Provides access to IPython to packages.

get_client() returns the Client object, from which you can construct a view.
This may return None if no client is available, if the user cancels, etc, in
which case you can raise an exception.

direct_view() returns a DirectView of the whole cluster. It is equivalent to
get_client()[:], with the difference that it will prompt the user to create new
engines if none are started.

load_balanced_view() returns a LoadBalancedView of the cluster. It is
equivalent to get_client().load_balanced_view(), with the difference that it
will prompt the user to create new engines if none are started.

parallel_map() might use IPython's parallel map_sync(), or the standard map()
function if IPython cannot be used.
If the 'ipython' keyword argument is True, the function will return an
additional boolean indicating whether this was computed through IPython (True)
or with the default map() function (False).

All of these functions have an 'ask' parameter, that indicates whether to
prompt the user to start a cluster if none is available. It is True by default
(except for parallel_map).
"""


from __future__ import division

__all__ = ['get_client', 'direct_view', 'load_balanced_view', 'parallel_map']


def get_client(ask=True):
    """Returns a Client object, from which you can construct a view.

    This may return None if no client is available, if the user cancels, etc.
    In this case, you might want to raise a ModuleError.
    """
    from engine_manager import EngineManager

    c = EngineManager.ensure_controller(connect_only=not ask)
    if c is not None and ask and not c.ids:
        EngineManager.start_engines(
                prompt="A module requested an IPython cluster, but no engines "
                       "are started. Do you want to start some?")
    if c is not None and c.ids:
        return c
    else:
        return None


def direct_view(ask=True):
    """Returns a DirectView of the whole cluster.

    This is equivalent to get_client()[:], with the difference that it will
    prompt the user to create new engines if none are started.
    """
    c = get_client()
    if c is not None:
        return c[:]
    else:
        return None


def load_balanced_view(ask=True):
    """Returns a LoadBalancedView of the cluster.

    This is equivalent to get_client().load_balanced_view(), with the
    difference that it will prompt the user to create new engines if none are
    started.
    """
    c = get_client()
    if c is not None:
        return c.load_balanced_view()
    else:
        return None


def parallel_map(function, *args, **kwargs):
    """Wrapper around IPython's map_sync() that defaults to map().

    This might use IPython's parallel map_sync(), or the standard map()
    function if IPython cannot be used.

    If the 'ask' keyword argument is true, the user will be prompted to start
    IPython engines, but the function will still default to map() if the user
    cancels.
    If the 'ipython' keyword argument is True, the function will return an
    additional boolean indicating whether this was computed through IPython
    (True) or with the default map() function (False).
    """
    say_ipython = kwargs.pop('ipython', False)
    ask = kwargs.pop('ask', False)
    if kwargs:
        raise TypeError("map() got unexpected keyword arguments")

    try:
        import IPython.parallel
    except ImportError:
        result, ipython = map(function, *args), False
    else:
        from engine_manager import EngineManager
        c = EngineManager.ensure_controller(connect_only=not ask)
        if c is not None and not c.ids:
            EngineManager.start_engines(
                    prompt="A module is performing a parallelizable "
                    "operation, however no IPython engines are running. Do "
                    "you want to start some?")

        if c is None or not c.ids:
            result, ipython = map(function, *args), False
        else:
            ldview = c.load_balanced_view()
            result, ipython = ldview.map_sync(function, *args), True

    if say_ipython:
        return result, ipython
    else:
        return result



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2013-2014, NYU-Poly.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import csv
import operator
try:
    import numpy
except ImportError: # pragma: no cover
    numpy = None

from ..common import TableObject, Table, InternalModuleError


def count_lines(fp):
    lines = 0
    for line in fp:
        lines += 1
    return lines


# FIXME : test coverage for CSVTable
class CSVTable(TableObject):
    def __init__(self, csv_file, header_present, delimiter,
                 skip_lines=0, dialect=None, use_sniffer=True):
        self._rows = None

        self.header_present = header_present
        self.delimiter = delimiter
        self.filename = csv_file
        self.skip_lines = skip_lines
        self.dialect = dialect

        (self.columns, self.names, self.delimiter,
         self.header_present, self.dialect) = \
            self.read_file(csv_file, delimiter, header_present, skip_lines,
                           dialect, use_sniffer)
        if self.header_present:
            self.skip_lines += 1

        self.column_cache = {}

    @staticmethod
    def read_file(filename, delimiter=None, header_present=True,
                  skip_lines=0, dialect=None, use_sniffer=True):
        if delimiter is None and use_sniffer is False:
            raise InternalModuleError("Must set delimiter if not using sniffer")

        try:
            with open(filename, 'rb') as fp:
                if use_sniffer:
                    first_lines = ""
                    line = fp.readline()
                    for i in xrange(skip_lines):
                        if not line:
                            break
                        line = fp.readline()
                    for i in xrange(5):
                        if not line:
                            break
                        first_lines += line
                        line = fp.readline()
                    sniffer = csv.Sniffer()
                    fp.seek(0)
                    if delimiter is None:
                        dialect = sniffer.sniff(first_lines)
                        delimiter = dialect.delimiter
                        # cannot determine header without sniffing delimiter
                        if header_present is None:
                            header_present = sniffer.has_header(first_lines)

                for i in xrange(skip_lines):
                    line = fp.readline()
                    if not line:
                        raise InternalModuleError("skip_lines greater than "
                                                  "the number of lines in the "
                                                  "file")

                if dialect is not None:
                    reader = csv.reader(fp, dialect=dialect)
                else:
                    reader = csv.reader(fp, delimiter=delimiter)
                result = reader.next()
                column_count = len(result)

                if header_present:
                    column_names = [name.strip() for name in result]
                else:
                    column_names = None
        except IOError:
            raise InternalModuleError("File does not exist")

        return column_count, column_names, delimiter, header_present, dialect

    def get_column(self, index, numeric=False):
        if (index, numeric) in self.column_cache:
            return self.column_cache[(index, numeric)]

        if numeric and numpy is not None:
            result = numpy.loadtxt(
                    self.filename,
                    dtype=numpy.float32,
                    delimiter=self.delimiter,
                    skiprows=self.skip_lines,
                    usecols=[index])
        else:
            with open(self.filename, 'rb') as fp:
                for i in xrange(self.skip_lines):
                    line = fp.readline()
                    if not line:
                        raise ValueError("skip_lines greater than the number "
                                         "of lines in the file")
                if self.dialect is not None:
                    reader = csv.reader(fp, dialect=self.dialect)
                else:
                    reader = csv.reader(fp, delimiter=self.delimiter)

                getter = operator.itemgetter(index)
                try:
                    result = []
                    for rownb, row in enumerate(reader, 1):
                        result.append(getter(row))
                except IndexError:
                    raise ValueError("Invalid CSV file: only %d fields on "
                                     "line %d (column %d requested)" % (
                                         len(row), rownb, index))
            if numeric:
                result = [float(e) for e in result]

        self.column_cache[(index, numeric)] = result
        return result

    @property
    def rows(self):
        if self._rows is not None:
            return self._rows
        with open(self.filename, 'rb') as fp:
            self._rows = count_lines(fp)
        self._rows -= self.skip_lines
        return self._rows


class CSVFile(Table):
    """Reads a table from a CSV file.

    This module uses Python's csv module to read a table from a file. It is
    able to guess the actual format of the file in most cases, or you can use
    the 'delimiter', 'header_present' and 'skip_lines' ports to force how the
    file will be read.
    """
    _input_ports = [
            ('file', '(org.vistrails.vistrails.basic:File)'),
            ('delimiter', '(org.vistrails.vistrails.basic:String)',
             {'optional': True}),
            ('header_present', '(org.vistrails.vistrails.basic:Boolean)',
             {'optional': True, 'defaults': "['True']"}),
            ('sniff_header', '(org.vistrails.vistrails.basic:Boolean)',
             {'optional': True, 'defaults': "['True']"}),
            ('skip_lines', '(org.vistrails.vistrails.basic:Integer)',
             {'optional': True, 'defaults': "['0']"}),
            ('dialect', '(org.vistrails.vistrails.basic:String)',
             {'optional': True})]
    _output_ports = [
            ('column_count', '(org.vistrails.vistrails.basic:Integer)'),
            ('column_names', '(org.vistrails.vistrails.basic:List)'),
            ('value', Table)]

    def compute(self):
        csv_file = self.get_input('file').name
        header_present = self.force_get_input('header_present', None)
        delimiter = self.force_get_input('delimiter', None)
        skip_lines = self.get_input('skip_lines')
        dialect = self.force_get_input('dialect', None)
        sniff_header = self.get_input('sniff_header')

        try:
            table = CSVTable(csv_file, header_present, delimiter, skip_lines,
                             dialect, sniff_header)
        except InternalModuleError, e:
            e.raise_module_error(self)

        self.set_output('column_count', table.columns)
        self.set_output('column_names', table.names)
        self.set_output('value', table)


_modules = [CSVFile]


###############################################################################

from StringIO import StringIO
import unittest
from vistrails.tests.utils import execute, intercept_result
from ..identifiers import identifier
from ..common import ExtractColumn


class CSVTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        import os
        cls._test_dir = os.path.join(
                os.path.dirname(__file__),
                os.pardir,
                'test_files')

    def test_csv_numeric(self):
        """Uses CSVFile and ExtractColumn to load a numeric array.
        """
        with intercept_result(ExtractColumn, 'value') as results:
            with intercept_result(CSVFile, 'column_count') as columns:
                self.assertFalse(execute([
                        ('read|CSVFile', identifier, [
                            ('file', [('File', self._test_dir + '/test.csv')]),
                        ]),
                        ('ExtractColumn', identifier, [
                            ('column_index', [('Integer', '1')]),
                            ('column_name', [('String', 'col 2')]),
                            ('numeric', [('Boolean', 'True')]),
                        ]),
                        ('PythonSource', 'org.vistrails.vistrails.basic', [
                            ('source', [('String', '')]),
                        ]),
                    ],
                    [
                        (0, 'value', 1, 'table'),
                        (1, 'value', 2, 'l'),
                    ],
                    add_port_specs=[
                        (2, 'input', 'l',
                         'org.vistrails.vistrails.basic:List'),
                    ]))
                # Here we use a PythonSource just to check that a numpy array
                # can be passed on a List port
        self.assertEqual(columns, [3])
        self.assertEqual(len(results), 1)
        self.assertEqual(list(results[0]), [2.0, 3.0, 14.5])

    def test_csv_mismatch(self):
        """Uses CSVFile and ExtractColumn with mismatching columns.
        """
        self.assertTrue(execute([
                ('read|CSVFile', identifier, [
                    ('file', [('File', self._test_dir + '/test.csv')]),
                ]),
                ('ExtractColumn', identifier, [
                    ('column_index', [('Integer', '0')]), # index is wrong
                    ('column_name', [('String', 'col 2')]),
                ]),
            ],
            [
                (0, 'value', 1, 'table'),
            ]))

    def test_csv_missing(self):
        """Uses CSVFile and ExtractColumn with a nonexisting column.
        """
        self.assertTrue(execute([
                ('read|CSVFile', identifier, [
                    ('file', [('File', self._test_dir + '/test.csv')]),
                ]),
                ('ExtractColumn', identifier, [
                    ('column_name', [('String', 'col not here')]),
                ]),
            ],
            [
                (0, 'value', 1, 'table'),
            ]))

    def test_csv_nonnumeric(self):
        """Uses CSVFile and ExtractColumn to load strings.
        """
        with intercept_result(ExtractColumn, 'value') as results:
            self.assertFalse(execute([
                    ('read|CSVFile', identifier, [
                        ('file', [('File', self._test_dir + '/test.csv')]),
                        ('header_present', [('Boolean', 'False')]),
                    ]),
                    ('ExtractColumn', identifier, [
                        ('column_index', [('Integer', '2')]),
                        ('numeric', [('Boolean', 'False')]),
                    ]),
                ],
                [
                    (0, 'value', 1, 'table'),
                ]))
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0],
                         ['col moutarde', '4', 'not a number', '7'])


class TestCountlines(unittest.TestCase):
    def test_countlines(self):
        # Simple
        fp = StringIO("first\nsecond")
        self.assertEqual(count_lines(fp), 2)

        # With newline at EOF
        fp = StringIO("first\nsecond\n")
        self.assertEqual(count_lines(fp), 2)

        # Empty
        fp = StringIO("")
        self.assertEqual(count_lines(fp), 0)

        # Single newline
        fp = StringIO("\n")
        self.assertEqual(count_lines(fp), 1)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import glob
import os
import sqlite3
from itertools import chain

from entity import Entity
from vistrail import VistrailEntity
from workflow import WorkflowEntity
from workflow_exec import WorkflowExecEntity
from thumbnail import ThumbnailEntity
from mashup import MashupEntity
from parameter_exploration import ParameterExplorationEntity

from vistrails.core.db.locator import FileLocator, BaseLocator
from vistrails.core.db.io import load_vistrail
import vistrails.core.system
import vistrails.db.services.io
from vistrails.core import debug

schema = ["create table entity(id integer primary key, type integer, "
          "name text, user integer, mod_time text, create_time text, "
          "size integer, description text, url text)",
          "create table entity_children(parent integer, child integer)",
          "create table entity_workspace(entity integer, workspace text)",
          "create table workspaces(id text primary key)",
          "insert into workspaces values ('Default')"]

class Collection(object):
    entity_types = dict((x.type_id, x)
                        for x in [VistrailEntity, WorkflowEntity, 
                                  WorkflowExecEntity, ThumbnailEntity,
                                  MashupEntity, ParameterExplorationEntity])
    def __init__(self, database=None):
        if database is None:
            self.database = ':memory:'
        else:
            self.database = database

        self.entities = {}
        self.deleted_entities = {}
        self.temp_entities = {}
        self.workspaces = {}
        self.currentWorkspace = 'Default'
        self.listeners = [] # listens for entity creation removal
        self.max_id = 0
        
        if not os.path.exists(self.database):
            debug.log("'%s' does not exist. Trying to create" % self.database)
            self.conn = sqlite3.connect(self.database)
            try:
                cur = self.conn.cursor()
                for s in schema:
                    cur.execute(s)
                self.conn.commit()
            except Exception, e:
                debug.critical("Could not create vistrail index schema", e)
        else:
            self.conn = sqlite3.connect(self.database)
        self.load_entities()

    #Singleton technique
    _instance = None

    @staticmethod
    def getInstance():
        if Collection._instance is False:
            debug.critical("Collection.getInstance() called but the "
                           "Collection has been deleted!")
            raise RuntimeError("Collection has been deleted!")
        elif Collection._instance is None:
            dotVistrails = vistrails.core.system.current_dot_vistrails()

            path = os.path.join(dotVistrails, "index.db")
            obj = Collection(path)
            Collection._instance = obj
        return Collection._instance

    @staticmethod
    def clearInstance():
        if Collection._instance:
            Collection._instance.conn.close()
        Collection._instance = False

    def add_listener(self, c):
        """ Add objects that listen to entity creation/removal
            Object may implement the following method
             updated(self)
        """
        self.listeners.append(c)

    def clear(self):
        cur = self.conn.cursor()
        cur.execute('delete from entity;')
        cur.execute('delete from entity_children;')
        cur.execute('delete from workspaces;')
        cur.execute('delete from entity_workspace;')

    def get_current_entities(self):
        """NOTE: returns an iterator"""
        return chain(self.entities.itervalues(), 
                     self.temp_entities.itervalues())

    def load_entities(self):
        cur = self.conn.cursor()
        cur.execute("select max(id) from entity;")
        for row in cur.fetchall():
            n = row[0]
            self.max_id = n if n is not None else 0
        
        cur.execute("select * from entity;")
        for row in cur.fetchall():
            entity = self.load_entity(*row)
            if entity is not None:
                self.entities[entity.id] = entity

        # now need to map children to correct places
        cur.execute("select * from entity_children;")
        for row in cur.fetchall():
            if row[0] in self.entities and row[1] in self.entities:
                self.entities[row[0]].children.append(self.entities[row[1]])
                self.entities[row[1]].parent = self.entities[row[0]]

        cur.execute("select * from workspaces;")
        for row in cur.fetchall():
            self.workspaces[row[0]] = []

        cur.execute("select * from entity_workspace;")
        for row in cur.fetchall():
            e_id, workspace = row
            if e_id in self.entities:
                if workspace not in self.workspaces:
                    self.workspaces[workspace] = []
                self.workspaces[workspace].append(self.entities[e_id])

    def save_entities(self):
        # TODO delete entities with no workspace
        for entity in self.deleted_entities.itervalues():
            self.db_delete_entity(entity)
        self.deleted_entities = {}
        
        for entity in self.entities.itervalues():
            if entity.was_updated:
                self.save_entity(entity)
                
        cur = self.conn.cursor()
        cur.execute('delete from workspaces;')
        cur.executemany("insert into workspaces values (?)", 
                        [(i,) for i in self.workspaces])

        cur.execute('delete from entity_workspace;')
        cur.executemany("insert into entity_workspace values (?, ?)", 
            ((k.id, i) for i,j in self.workspaces.iteritems() for k in j))

    def load_entity(self, *args):
        if args[1] in Collection.entity_types:
            entity = Collection.entity_types[args[1]].create(*args)
            return entity
        else:
            debug.critical("Cannot find entity type '%s'" % args[1])
        return None

    def add_entity(self, entity):
        """ Adds entities to memory recursively """
        if entity.id is None:
            self.max_id += 1
            entity.id = self.max_id
        entity.was_updated = True
        self.entities[entity.id] = entity
        for child in entity.children:
            child.parent = entity
            self.add_entity(child)
        
    def add_temp_entity(self, entity):
        """Add an entity to memory only.  Used for vistrails that have
        not been saved."""
        self.max_id += 1
        entity.id = self.max_id
        self.temp_entities[entity.id] = entity
        for child in entity.children:
            child.parent = entity
            self.add_temp_entity(child)

    def is_temp_entity(self, entity):
        return entity.id in self.temp_entities
            
    def save_entity(self, entity):
        """ saves entities to disk """
        cur = self.conn.cursor()
        entity_tuple = entity.save()
        cur.execute("insert or replace into entity values(%s)" % \
                    ','.join(('?',) * len(entity_tuple)), entity_tuple)
        entity.was_updated = False
        cur.execute('delete from entity_children where parent=?', (entity.id,))
        cur.executemany("insert into entity_children values (?, ?)",
                        ((entity.id, child.id) for child in entity.children))

    def commit(self):
        self.save_entities()
        self.conn.commit()
        for o in self.listeners:
            if hasattr(o, "updated"):
                o.updated()

    def delete_entity(self, entity):
        """ Delete entity from memory recursively """
        if entity.id in self.temp_entities:
            del self.temp_entities[entity.id]
        else:
            self.deleted_entities[entity.id] = entity
            if entity.id in self.entities:
                del self.entities[entity.id]
        for child in entity.children:
            self.delete_entity(child)

    def add_workspace(self, workspace):
        if workspace not in self.workspaces:
            self.workspaces[workspace] = []

    def add_to_workspace(self, entity, workspace=None):
        if not entity:
            return
        if not workspace:
            workspace = self.currentWorkspace
        self.add_workspace(workspace)
        if entity not in self.workspaces[workspace]:
            self.workspaces[workspace].append(entity)
        
    def del_from_workspace(self, entity, workspace=None):
        if not workspace:
            workspace = self.currentWorkspace
        if workspace in self.workspaces:
            if entity in self.workspaces[workspace]:
                self.workspaces[workspace].remove(entity)

    def delete_workspace(self, workspace):
        if workspace in self.workspaces:
            del self.workspaces[workspace]

    def db_delete_entity(self, entity):
        """ Delete entity from database """
        cur = self.conn.cursor()
        if entity.id is not None:
            cur.execute("delete from entity where id=?", (entity.id,))
            cur.execute("delete from entity_children where parent=?", (entity.id,))
            cur.execute("delete from entity_children where child=?", (entity.id,))

    def create_workflow_entity(self, workflow):
        entity = WorkflowEntity(workflow)
        self.add_entity(entity)
        return entity

    def create_vistrail_entity(self, vistrail):
        entity = VistrailEntity(vistrail)
        self.add_entity(entity)
        return entity

    def update_from_directory(self, directory):
        filenames = glob.glob(os.path.join(directory, '*.vt'))
        for filename in filenames:
            locator = FileLocator(filename)
            url = locator.to_url()
            self.updateVistrail(url)

    def fromUrl(self, url):
        """ Check if entity with this url exist in index and return it """
        for e in self.entities.itervalues():
            if e.url == url:
                return e
        return None

    def urlExists(self, url):
        """ Check if entity with this url exist """
        locator = BaseLocator.from_url(url)
        if locator.is_valid():
            return True
        return False

    def updateVistrail(self, url, vistrail=None):
        """ updateVistrail(self, string:url, Vistrail:vistrail)
        Update the specified entity url. Delete or reload as necessary.
        Need to make sure workspaces are updated if the entity is changed.
        """
        entities = [e for e in self.entities.itervalues() if e.url == url]
        entity = entities[0] if len(entities) else None
        while entity and entity.parent:
            entity = entity.parent 
            url = entity.url
        workspaces = [p for p in self.workspaces if entity in self.workspaces[p]]
        if entity:
            for p in workspaces:
                self.del_from_workspace(entity, p)
            self.delete_entity(entity)

        locator = BaseLocator.from_url(url)
        if locator.is_valid():
            if not vistrail:
                (vistrail, abstractions, thumbnails, mashups) = load_vistrail(locator)
                vistrail.abstractions = abstractions
                vistrail.thumbnails = thumbnails
                vistrail.mashups = mashups
            entity = self.create_vistrail_entity(vistrail)
            for p in workspaces:
                self.add_to_workspace(entity, p)
            return entity
        else:
            # probably an unsaved vistrail
            pass
#            debug.critical("Locator is not valid!")



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core import debug
import copy

##############################################################################

class InternalTuple(object):
    """Tuple used internally for constant tuples."""

    list_depth = 0

    def _get_length(self, length):
        return len(self._values)
    def _set_length(self, length):
        self._values = [None] * length
    length = property(_get_length, _set_length)

    def compute(self):
        return

    def set_input_port(self, index, connector):
        self._values[index] = connector()

    def get_output(self, port):
        return tuple(self._values)

    def update(self):
        pass

##############################################################################

class AbortExecution(Exception):
    """Internal exception raised to signal the interpreter it should stop.
    """

##############################################################################

class BaseInterpreter(object):

    def __init__(self):
        """ BaseInterpreter() -> BaseInterpreter
        Initialize class members
        
        """
        self.done_summon_hook = None
        self.done_update_hook = None

    def get_name_dependencies(self, astList):
        """get_name_dependencies(astList) -> list of something 
        
        """
        
        result = []
        if astList[0]==1: # NAME token
            result += [astList[1]]
        else:
            for e in astList:
                if isinstance(e, list):
                    result += self.get_name_dependencies(e)
        return result

#    def build_alias_dictionary(self, pipeline):
#        aliases = {}
#        for mid in pipeline.modules:
#            for f in pipeline.modules[mid].functions:
#                fsig = f.getSignature()
#                for pidx in xrange(len(f.params)):
#                    palias = f.params[pidx].alias
#                    if palias and palias!='':
#                        for f1 in reversed(pipeline.modules[mid].functions):
#                            if f1.getSignature()==fsig:
#                                p = f1.params[pidx]
#                                aliases[palias] = (p.type, expression.parse_expression(str(p.strValue)))
#                                break
#        return aliases

    def compute_evaluation_order(self, aliases):
        # Build the dependencies graph
        dp = {}
        for alias,(atype,(base,exp)) in aliases.items():
            edges = []
            for e in exp:
                edges += self.get_name_dependencies()
            dp[alias] = edges
            
        # Topological Sort to find the order to compute aliases
        # Just a slow implementation, O(n^3)...
        unordered = copy.copy(list(aliases.keys()))
        ordered = []
        while unordered:
            added = []
            for i in xrange(len(unordered)):
                ok = True
                u = unordered[i]
                for j in xrange(len(unordered)):
                    if i!=j:
                        for v in dp[unordered[j]]:
                            if u==v:
                                ok = False
                                break
                        if not ok: break
                if ok: added.append(i)
            if not added:
                debug.warning('Looping dependencies detected!')
                break
            for i in reversed(added):
                ordered.append(unordered[i])
                del unordered[i]
        return ordered

    def evaluate_exp(self, atype, base, exps, aliases):
        # FIXME: eval should pretty much never be used
        import datetime
        for e in exps:
            base = (
                    base[:e[0]] + str(eval(
                        e[1],
                        {'datetime': locals()['datetime']},
                        aliases)) +
                    base[e[0]:])
        if not atype in ['string', 'String']:
            if base=='':
                base = '0'
            try:
                base = eval(base)
            except Exception:
                pass
        return base

    def resolve_aliases(self, pipeline,
                        customAliases=None):
        # We don't build the alias dictionary anymore because as we don't 
        # perform expression evaluation anymore, the values won't change.
        # We only care for custom aliases because they might have a value 
        # different from what it's stored.
        
        aliases = {}
        if customAliases:
            #customAliases can be only a subset of the aliases
            #so we need to build the Alias Dictionary always
            for k,v in customAliases.iteritems():
                aliases[k] = v
            # no support for expression evaluation. The code that does that is
            # ugly and dangerous.
#        ordered = self.compute_evaluation_order(aliases)
#        casting = {'int': int, 'float': float, 'double': float, 'string': str,
#                   'Integer': int, 'Float': float, 'String': str}
#        for alias in reversed(ordered):
#            (atype,base) = aliases[alias]
#            #no expression evaluation anymore
#            aliases[alias] = base
#            #value = self.evaluate_exp(atype,base,exps,aliases)
#            #aliases[alias] = value
        for alias in aliases:
            try:
                info = pipeline.aliases[alias]
                param = pipeline.db_get_object(info[0],info[1])
                param.strValue = str(aliases[alias])
            except KeyError:
                pass
                    
        return aliases
    
    def update_params(self, pipeline,
                        customParams=None):
        """update_params(pipeline: Pipeline, 
                         customParams=[(vttype, oId, strval)] -> None
        This will set the new parameter values in the pipeline before
        execution 
        
        """
        if customParams:
            for (vttype, oId, strval) in customParams:
                try:
                    param = pipeline.db_get_object(vttype,oId)
                    param.strValue = str(strval)
                except Exception, e:
                    debug.debug("Problem when updating params", e)

    def resolve_variables(self, vistrail_variables, pipeline):
        for m in pipeline.module_list:
            if m.is_vistrail_var():
                vistrail_var = vistrail_variables(m.get_vistrail_var())
                if vistrail_var is None: # assume set in parameter exploration
                    continue
                strValue = vistrail_var.value
                for func in m.functions:
                    if func.name == 'value':
                        func.params[0].strValue = strValue

    def set_done_summon_hook(self, hook):
        """ set_done_summon_hook(hook: function(pipeline, objects)) -> None
        Assign a function to call right after every objects has been
        summoned during execution
        
        """
        self.done_summon_hook = hook

    def set_done_update_hook(self, hook):
        """ set_done_update_hook(hook: function(pipeline, objects)) -> None
        Assign a function to call right after every objects has been
        updated
        
        """
        self.done_update_hook = hook

##############################################################################



#!/usr/bin/env python

# This example demonstrates the use of VTK data arrays as attribute
# data as well as field data. It creates geometry (vtkPolyData) as
# well as attribute data explicitly.

import vtk

# Create a float array which represents the points.
pcoords = vtk.vtkFloatArray()
# Note that by default, an array has 1 component.
# We have to change it to 3 for points
pcoords.SetNumberOfComponents(3)
# We ask pcoords to allocate room for at least 4 tuples
# and set the number of tuples to 4.
pcoords.SetNumberOfTuples(4)
# Assign each tuple. There are 5 specialized versions of SetTuple:
# SetTuple1 SetTuple2 SetTuple3 SetTuple4 SetTuple9
# These take 1, 2, 3, 4 and 9 components respectively.
pcoords.SetTuple3(0, 0.0, 0.0, 0.0)
pcoords.SetTuple3(1, 0.0, 1.0, 0.0)
pcoords.SetTuple3(2, 1.0, 0.0, 0.0)
pcoords.SetTuple3(3, 1.0, 1.0, 0.0)

# Create vtkPoints and assign pcoords as the internal data array.
points = vtk.vtkPoints()
points.SetData(pcoords)

# Create the cells. In this case, a triangle strip with 2 triangles
# (which can be represented by 4 points)
strips = vtk.vtkCellArray()
strips.InsertNextCell(4)
strips.InsertCellPoint(0)
strips.InsertCellPoint(1)
strips.InsertCellPoint(2)
strips.InsertCellPoint(3)

# Create an integer array with 4 tuples. Note that when using
# InsertNextValue (or InsertNextTuple1 which is equivalent in
# this situation), the array will expand automatically
temperature = vtk.vtkIntArray()
temperature.SetName("Temperature")
temperature.InsertNextValue(10)
temperature.InsertNextValue(20)
temperature.InsertNextValue(30)
temperature.InsertNextValue(40)

# Create a double array.
vorticity = vtk.vtkDoubleArray()
vorticity.SetName("Vorticity")
vorticity.InsertNextValue(2.7)
vorticity.InsertNextValue(4.1)
vorticity.InsertNextValue(5.3)
vorticity.InsertNextValue(3.4)

# Create the dataset. In this case, we create a vtkPolyData
polydata = vtk.vtkPolyData()
# Assign points and cells
polydata.SetPoints(points)
polydata.SetStrips(strips)
# Assign scalars
polydata.GetPointData().SetScalars(temperature)
# Add the vorticity array. In this example, this field
# is not used.
polydata.GetPointData().AddArray(vorticity)

# Create the mapper and set the appropriate scalar range
# (default is (0,1)
mapper = vtk.vtkPolyDataMapper()
mapper.SetInput(polydata)
mapper.SetScalarRange(0, 40)

# Create an actor.
actor = vtk.vtkActor()
actor.SetMapper(mapper)

# Create the rendering objects.
ren = vtk.vtkRenderer()
ren.AddActor(actor)

renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)

iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

iren.Initialize()
renWin.Render()
iren.Start()
